@@ -565,10 +565,6 @@ destroy_header_table (smtp_message_t message)
 
   assert (message != NULL);
 
-  /* Take out the hash table */
-  h_destroy (message->hdr_action, NULL, NULL);
-  message->hdr_action = NULL;
-
   /* Take out the linked list */
   for (header = message->headers; header!= NULL; header = next)
     {
@@ -578,6 +574,14 @@ destroy_header_table (smtp_message_t message)
       free (header->header);
       free (header);
     }
+
+  /* Take out the hash table */
+  if (message->hdr_action != NULL)
+    {
+      h_destroy (message->hdr_action, NULL, NULL);
+      message->hdr_action = NULL;
+    }
+
   message->headers = message->end_headers = NULL;
 }
 
@@ -589,6 +593,10 @@ find_header (smtp_message_t message, const char *name, int len)
 
   assert (message != NULL && name != NULL);
 
+  if (len < 0)
+    len = strlen (name);
+  if (len == 0)
+    return NULL;
   info = h_search (message->hdr_action, name, len);
   if (info == NULL && (p = memchr (name, '-', len)) != NULL)
     info = h_search (message->hdr_action, name, p - name + 1);
@@ -670,7 +678,8 @@ process_header (smtp_message_t message, const char *header, int *len)
 
   assert (message != NULL && header != NULL && len != NULL);
 
-  if ((p = memchr (header, ':', *len)) != NULL
+  if (*len > 0
+      && (p = memchr (header, ':', *len)) != NULL
       && (info = find_header (message, header, p - header)) != NULL)
     {
       if ((action = info->action) != NULL)
@@ -842,30 +851,33 @@ smtp_set_header_option (smtp_message_t message, const char *header,
       return 0;
     }
 
-  if (info->prohibit || (info->action->flags & (PROHIBIT | PRESERVE)))
+  /* Don't permit options to be set on headers that must pass intact or
+     which are prohibited. */
+  if (info->action->flags & (PROHIBIT | PRESERVE))
     {
       set_error (SMTP_ERR_INVAL);
       return 0;
     }
 
-  if (option == Hdr_OVERRIDE)
+  /* There is an odd quirk when setting options.  Setting an option for
+     the OPTIONAL headers known to libESMTP causes default values to be
+     generated automatically when not found in the message, so long as
+     there is no other reason to prevent them appearing in the message! */
+
+  /* Don't allow the user to set override on prohibited headers. */
+  if (option == Hdr_OVERRIDE && !info->prohibit)
     {
-      /* There is an odd quirk of the Hdr_OVERRIDE option.  Setting it
-         to false for the OPTIONAL headers known to libESMTP causes
-         default values to be generated automatically when not found in
-         the message! */
       va_start (alist, option);
       info->override = !!va_arg (alist, int);
       va_end (alist);
       return 1;
     }
-  if (option == Hdr_PROHIBIT)
+  /* Don't allow the user to prohibit required headers. */
+  if (option == Hdr_PROHIBIT && !(info->action->flags & REQUIRE))
     {
-      /* Can't set prohibit if a header of this name already set up
-         or the header is required. */
-      if (info->hdr != NULL || (info->action->flags & REQUIRE))
-        return 0;
-      info->prohibit = 1;
+      va_start (alist, option);
+      info->prohibit = !!va_arg (alist, int);
+      va_end (alist);
       return 1;
     }
 
